/*** Giddes
 *     - Graphical Interface for Designing Discrete Event Systems
 *  S.Laurie Ricker Ph.D
 *  
 *  Developed by I. Fraser & W. Fowler 
 *           
 *  
 * reproduction without the express permission of Toba will
 * result in being taken outside and shot many times with a 
 * very tiny gun.
 ***/

//*********************************************************
// Documentation completed January 18, 2002               *
//*********************************************************

/*	
	This class is the bucket that is comnbined to make the BucketList
	A bucket has a name and a collection of transitions. The bucket's
	name is the same as the originating state of all the transitions
	in the collection. 
	
	All the transitions are sorted upon entry to the collection so
	any particular edge can be found easily in the bucket with a 
	binary search. All the ordering of the collection is handled 
	by the bucket itself.
	
	The bucket can also return the entire collection with getTransitions()
*/

/******************************************************************
/ Optimization Queue:  1) String name should be replaced by State *
/                      2) Binary Searches need to be implemented  *
/                                                                 *
/                                                                 *
/*****************************************************************/


package umdes;

import java.util.Vector;
import java.util.Enumeration;

/**
 * A Bucket is a specialized collection designed for strictly Transitions.
 * The Transitions are sorted internally. Several Buckets are used in the BucketList collection
 * to increase ADT efficency for the Transitions.
 *
 * @see UmdesAutomaton
 * @see Transition
 * @see BucketList
 */
public class Bucket{
	//*************************************************************
	// protected members
	//*************************************************************
	/**
	 * The collection of Transitions in the Bucket. Vector chosen because, many transitions
	 * is not necessarily likely and so extra overhead of HashTable is not justified for this 
	 * Implementation.
	 */
	protected Vector<Transition> transitions;
	/** 
   * The name of the Bucket
   */
	//protected String name; //<--- THIS WILL BE CHANGED TO STATE AND METHODS NEED
	                       // TO BE UPDTATED TO REFLECT THIS CHANGE
	protected State name;
	
	//*************************************************************
	// public members
	//*************************************************************
  
  /**
   * Creates a Bucket with the name n
   *
   * @param n the name of the new bucket
   *
   * @throws an Invalid State Name Exception if the name given is 
   * not a valid state name
   */
 	public Bucket(String n) throws Exception{
	  transitions = new Vector<Transition>();
	  name=new State(n);
	}
	
  /**
   * Creates a Bucket with the the State s
   *
   * @param s the State of the new bucket
   */
  public Bucket(State s){
    transitions = new Vector<Transition>();
    name=new State(s);
  }
  
  /**
   * Copy Constructor, creates a copy of the Bucket b
   *
   * @param b the Bucket to be copied
   */
  public Bucket(Bucket b){
    Transition t;
    
	transitions=new Vector<Transition>();
	name =new State(b.getState());
	
	Enumeration<Transition> e=b.getTransitions().elements();
    while(e.hasMoreElements()){
      t = e.nextElement();
      addTransition(t);
    }
		
		
		  
	}
	
	/**
	 * Generate this Bucket's hashcode
	 *
	 * @return int hash code for this Bucket
	 */ 
  public int hashCode(){
    int code=name.hashCode();
    Enumeration<Transition> e=transitions.elements();
    Transition t;
    while(e.hasMoreElements()){
      t=(Transition)e.nextElement();
      code+=t.hashCode();
    }
    
    return code;
  }
  
	/**
	 * Clone the Bucket
	 */
	public Object clone(){
	  Bucket b= new Bucket(this);
	  return(b);
  }
  /**
   * Clears the bucket of all of its contents
   */
 	public void clearBucket(){
	  transitions.clear();
	  name=null;
	}
	
  /**
   * Accessor to the name of the Bucket and of its State
   *
   * @return Name of the bucket
   */
	public String getName(){
	  return(name.toString());
	}
	
	/**
	 * Accessor to the State of the Bucket
	 *
	 * @return the Bucket's State
	 */
	public State getState(){
	  return((State)name.clone());
  }
   
  /**
   * Sets the name of the Bucket. If the name is NOT
   * a valid state name the bucket's name is not changed
   * Note: the buckets contents are not changed, to change
   * all the buckets contents to reflect the name use 
   * renameBucket()
   *
   * @param n The name of the bucket
   */
	public void setName(String n){
	  State temp;
	  try{
	   temp=new State(n);
	  }
	  catch(Exception e){
	    return;
	  }
	  
	  // set bucket's state
	  name=temp;
	}
	
	/**
	 * Sets the state of the Bucket
	 * Note: the buckets contents are not changed 
	 * all the buckets contents to reflect the name use 
	 * renameBucket()
	 *
	 * @param s the new State for the Bucket
	 */
	public void setState(State s){
	  name=(State)s.clone(); // make copy to make sure that references don't get bundled
	}
	
  /**
   * Tests see if there is at least one transition
   * in the Bucket with the RegExpr r as a label
   *
   * @param r the label to look for
   *
   * @return True if the exists a Transition with the label r in the Bucket
   */
	public boolean hasTransitionWith(RegExpr r){
	  int i;

	  Transition temp;
	  
	  for(i=0;i<transitions.size();i++){
	  	temp= (Transition)transitions.elementAt(i);
	  	// check to see if r is too small can't be in the bucket
	  	if(r.isLessThan(temp.getLabel()))
	  		return(false);
	  	// check to see if you have found the transition
	    if(r.equals(temp.getLabel()))
		  return(true);
	  }  
	  	
	  return(false);	
	}	
	
  /**
   * Returns the collection of transitions with the RegExpr r
   * as their label that in this Bucket.
   *
   * @param r - The label to be checked
   *
   * @return The collection of Transitions if any are found.
   * If the transition is not found the vector will be empty.
   */
	public Vector<Transition> hasTransition(RegExpr r){
	  int i;
	  
	  Vector<Transition> index = new Vector<Transition>();
	  Transition temp;
	  
	  for(i=0;i<transitions.size();i++){
	  	temp= (Transition)transitions.elementAt(i);
	  	
	  	// check to see if r is too small can't be in the bucket
	  	if(r.isLessThan(temp.getLabel()))
	  		return(index);
	  	// check to see if you have found the transition
	    if(r.equals(temp.getLabel()))
		  index.add(temp);
	  }  
	  	
	  return(index);	
	}	
	
  /**
   * Returns the index of the Transition if it is found in the Bucket.
   * A -1 return indicates that no such Transition exists.
   * Note: This function may be deprecated or made protected.
   *
   * @param r - The label to be checked in the Transitions
   * @param st - The To State to be checked in the Transitions
   *
   * @return -1 If the transition is not found.
   * otherwise an int > 0  where int is the index of the first Transition found.
   */
 	public int hasTransition(RegExpr r, State st){
	  int i;
	  Transition temp;
	  
	  for(i=0;i<transitions.size();i++){
	  	temp= (Transition)transitions.elementAt(i);
	  	
	  	// check to see if r is too small can't be in the bucket
	  	if(r.isLessThan(temp.getLabel()))
	  		return(-1);
	  	// check to see if you have found the transition
	    if(r.equals(temp.getLabel())&& st.equals(temp.getToState()))
		  return(i);
		  
	  }  
	  return(-1);	
	}	 
	  	

  /**
   * Adds Transition t to the Bucket if it is not already there.
   * All Transitions added to the Bucket must have the same originating State
   * The name of the Originating State must also be name of the Bucket
   * Only a Transition that has the an appropriate Originating State and is not
   * already in the Bucket will be added. 
   *
   * @see Transition
   *
   * @param t the Transition to add to the Bucket
   *
   * @return True if the Transition is successfully added
   */
 	public boolean addTransition(Transition t){
	  int i;
	  Transition temp;
	  RegExpr r = t.getLabel();
	  
	  // check to see if the Transition has the right Originating State 
	  // for this Bucket
    if(!t.getFromState().equals(name)){
      return(false);
    }
	  // trivial case where the bucket is empty (no search required)
	  if(isEmpty()){
	    transitions.add(t);		 
		  return(true);
		}
	  
	  for(i=0;i<transitions.size();i++){
	  	temp= (Transition)transitions.elementAt(i);
	  	
	  	// if the transition is already there
	  	if(temp.equals(t))
	  	  return(false);
	  	
	    // if the appropiate spot is found 
	    if(r.isLessThan(temp.getLabel())){
	      // insert at index i
	      transitions.add(i,t);		 
		  return(true);
	    }
	  }
	  // the transition is to be appended
	  transitions.add(t);		 
	  return(true);
	}
	
  /**
   * Removes the specified Transition from the bucket if it is there
   *
   * @param t the Transition to be removed
   * @return True if the Transition is found and removed
   */
 	public boolean removeTransition(Transition t){
	  int result;
		
		result=hasTransition(t.getLabel(),t.getToState());
  
		// not found
		if(result==-1){
		  return(false);
		}
		//remove
		transitions.removeElementAt(result);
		return(true);
	}
	
  /**
   * Tests whether Object b is equal to the Bucket. Returns True if the Object b is a Bucket
   * and it has the same transitions as the Bucket 
   *
   * @return True if the Object b (is a Bucket and) has the same transitions 
   * as this bucket
   */
 	public boolean equals(Object b){
		int i;
		Transition t1,t2=null;
		Enumeration<Transition> e;
		boolean found=false;
		
		if(this==b)  // same instance
		  return(true);
		
		if(b==null){
  	  return(false);
		}
		
		Bucket bkt;
		try{
		  bkt=(Bucket)b;
		}
		catch(ClassCastException cce){
		  return(false);
		}
		
		if(size()!=bkt.size()){
  	  return(false);
		}
		
		// check each transition to see if it is in the bucket
		// remember order is not guarranteed by Enumeration!
		for(i=0;i<transitions.size();i++){
		  t1=(Transition)transitions.elementAt(i);  
		  found=false;
		  e=bkt.transitions.elements();
		  while(e.hasMoreElements()){
		    t2 = e.nextElement();
		    if(t1.equals(t2)){
		      found=true;
		      break; // get out of second loop because t1 is matched to t2
		    }
		  }
		  if(!found){
		    return(false);
		  }
		}
		return(true);
	}
	
  /**
   * Accessor function to all of the Transitions in the Bucket
   *
   * @return The collection of transitions in the bucket
   */
 	public Vector<Transition> getTransitions(){
	  Vector<Transition> v;
	  v=new Vector<Transition>(transitions);
	  return(v);
	}
	
  /**
   * Accesssor function to all of the Bucket's Transitions
   *  with the label r
   *
   * @param  r the label to search for
   *
   * @return the collection of transitions found with the label r
   * null return if none are found 
   */
 	public Vector<Transition> getTransition(RegExpr r){
		Vector<Transition> trans;

		trans = hasTransition(r);
		if(trans.size()>0)		
		  return(trans);
		else
		  return(null);
	}
  /**
   * Tests for an empty Bucket
   *
   * @return True if the Bucket is empty
   */
 	public boolean isEmpty(){
	  return(transitions.isEmpty());
	}
	
  /**
   * Gets the size of the Bucket
   * 
   * @return int - The number of transitions in the Bucket
   */
  public int size(){
	  return(transitions.size());
	}
	
  /**
   * Returns a Vector of the States that are
   * visited from the State st with an event RegExpr r 
   * Note: All Transitions in this bucket should start at State st
   *
   * @param st the starting state 
   * @param r the event that occured
   *
   * @return the collection of states that can be reached by this event
   * null returned if no states can be reached with r
   */
 	public Vector<State> doTransition(State state,RegExpr r){
 		
	  Vector<State> result = new Vector<State>();
	  Vector<Transition> trans = getTransition(r);
	  
	  if(trans==null) return(null);
	  else{
		  for(int i = 0; i < trans.size() ; i++){
			  Transition t = trans.elementAt(i);
			  result.addElement(new State(t.doTransition(state,r)));
		  }
	  }
	  
	  return(result);
	}
	
  /**
   * Renames an existing Transition in the Bucket with new properties
   * If the Transition t is not present, method returns without modifying the Bucket
   *
   * @param t the existing transition to be modified
   * @param newTo the new state the transition is to go to
   * @param newLabel the new label of the transition 
   *
   */
	public void editTransition(Transition t,State newTo,RegExpr newLabel){
	 if(hasTransition(t.getLabel(),t.getToState())<0)
	   return;

	 Transition tempTrans=new Transition(t.getFromState(),newTo,newLabel); // the new transition
	 
     // switch the transitions 
     removeTransition(t);
     addTransition(tempTrans);
  }
  
  /**
   * Renames the bucket and all the Originating States
   * of the Transitions in the Bucket.
   * This changes the entire Bucket unlike setState() and setName()
   *
   * @param from the new Originating States and the new name for the bucket
   */
  public void renameBucket(State from){
    int i;
    Transition t;
    
    // rename the bucket itself
    name=(State)from.clone();
    
    // change all the transitions
    for(i=0;i<transitions.size();i++){
      t=(Transition)transitions.elementAt(i);
      t.setFromState(from);
      transitions.setElementAt(t,i);
    }
  }
  
  /**
   * Returns true if the Bucket contains an Epsilon Transition
   *
   * @see Transition
   *
   * @return True if the Bucket contains an Epsilon Transition
   */
  public boolean hasETransition(){
    Enumeration<Transition> e;
    Transition t;
    
    // check for an epsilon Transition
    e=transitions.elements();
    while(e.hasMoreElements()){
      t = e.nextElement();
      if(t.isEpsilon())
        return(true);
    }
    //default return
    return(false);
  }
    
  /**
   * Checks to see if Bucket has non-deterministic Transitions
   * For non-deterministic Transitions see Glossary
   *
   * @return True if the Bucket has non-deterministic Transitions
   */
  public boolean isNondeter(){
    Enumeration<Transition> e;
    Transition t;
    Vector<TransitionLabel> temp= new Vector<TransitionLabel>();
    
    e=transitions.elements();
    while(e.hasMoreElements()){
      t= e.nextElement();
      
      if(t.isEpsilon())
        return(true);
           
      if(temp.contains(t.getLabel())) // already one found
        return(true);
      
      temp.add(t.getLabel());// add the label to collection of those found
    }
    return(false);
  }
  
  /**
   * Returns True if there Exists a Transition in this Bucket that goes to 
   * State b
   *
   * @param b the State to checked for
   *
   * @return True if there Exists a Transition in this Bucket that goes to 
   * State b
   */
  public boolean isToB(State b){  
	  Enumeration<Transition> trans = transitions.elements();
	  while(trans.hasMoreElements()){
		  Transition t = trans.nextElement();
		  if(t.goesToState(b)) return true;
	  }
	  return false;
  }        
  
};
