/*** Giddes
 *     - Graphical Interface for Designing Discrete Event Systems
 *  S.Laurie Ricker Ph.D
 *
 *  Developed by I. Fraser & W. Fowler
 *
 *
 * reproduction without the express permission of Toba will
 * result in being taken outside and shot many times with a
 * very tiny gun.
 ***/

//*********************************************************
// Documentation completed January 17, 2002               *
//*********************************************************


package umdes;
import java.util.Vector;
/**
 * This Object describes a regular expression that can be used as event label
 * in a Transition. It supports all usual operations with regular expression as expected
 * The Constructors only accept valid regular expression syntax. Any malformed expressions
 * will be rejected as described in the method documentation.
 *
 * @see UmdesAutomaton
 * @see Transition
 */
public class RegExpr
{

	//*******************************************************
	// constants
	//*******************************************************

	/**
	 * This represents an empty transition, specified with a greek epsilon in usual notation
	 * Note: In the beta version this an 'e'. This will be changed for the first version.
	 */
  public static final RegExpr EPSILON = new RegExpr("e"); //\u03B5" unicode epsilon
  /**
   * This represents an null transition, specified with 0 as in usual mathematical notation
   */
  public static final RegExpr NULL = new RegExpr("@"); //\u1D7D8 unicode Null
  /**
   * This is a open bracket that can be used to build more complex regular expressions
   */
  public static final String OPEN_BRACKET="(";
  /**
   * This is a close bracket that can be used to build more complex regular expressions
   */
  public static final String CLOSE_BRACKET=")";
  /**
   * This is the union operator that can be used to build more complex regular expressions
   */
  public static final String UNION="+";
  /**
   * This is the Kleen star operator that can be used to build more complex regular expressions
   */
  public static final String STAR="*";


  //public static final String CONCAT="\u2022";  // unicode dot for concatenation

  //***********************************************************
  // protected members
	//***********************************************************

	/**
	 * This is the regular expression
	 */
  protected String expression;

  //************************************************************
  // public members
	//************************************************************

  /**
   * Creates a new RegExpr if String s is a valid expression.
   * If s is not a regular expression the new RegExpr will be empty.
   *
   * @param String - This is the expression that is to be converted
   * into a regualar expression data type
   */
  public RegExpr(String s){
 	  if(isRegExpression(s))
	    expression=s;
	  else
	    expression= new String();
  }

  /**
   * Creates an Empty RegExpr
   */
  public RegExpr(){
    expression=new String();
  }


  /**
   * Creates a new RegExpr if StringBuffer s is a valid expression.
   * If s is not a regular expression the new RegExpr will be empty.
   *
   * @param String - This is the expression that is to be converted
   * into a regualar expression data type
   */
  public RegExpr(StringBuffer s){
 	  if(isRegExpression(s.toString()))
	   expression=s.toString();
	  else
	   expression= new String();
  }

  /**
   * Copy Constructor for a RegExpr. An alternative is to use clone()
   *
   * @param RegExpr r - the Regular Expression to copy
   */
  public RegExpr(RegExpr r){
  	expression=new String(r.getExpression());
  }

  public Object clone(){
    return(new String(expression));
  }
  /**
   * Accessor function for the expression
   *
   * @return The expression of the Object
   */
  public String getExpression(){
  	return(new String(expression));
  }

  /**
   * Tests whether this RegExpr is an Epsilon
   *
   * @return True if this RegExpr is an epsilon
   */
  public boolean isEpsilon(){
    return(expression.equals(EPSILON.getExpression()));
  }

  /**
   * Takes the Union of r and this RegExpr. Brackets are inserted to maintain precedence
   * i.e. RegExpr r1.union(r2) becomes a RegExpr with the expression (r1)+(r2)
   *
   * @param r A regular expression to Union with this expression
   */
  public void union(RegExpr r){
  	// perform union with + as the operator
  	// ie: insert brackets and plus so expression is now
  	// (expression)+(r)
  	String result;

   	result = new String(OPEN_BRACKET);

  	result=result.concat(expression);

    result=result.concat(CLOSE_BRACKET); // close bracket
  	result=result.concat(UNION); // operator
  	result=result.concat(OPEN_BRACKET); // open bracket
  	result=result.concat(r.getExpression());
  	result=result.concat(CLOSE_BRACKET); // close bracket

  	// set new expression
  	expression=result;
  }

  /**
   * Takes the Union of String s and the RegExpr if s is a valid regular expression
   * This a convenient overload of the usual union(RegExpr) function
   *
   * @param s The String to union with the RegExpr. It must be a valid regular expresssion
   * @return True if union was successful
   */
  public boolean union(String s){

  	if(isRegExpression(s)){
  	  // perform union with + as the operator
  	  RegExpr temp= new RegExpr(s);
  	  union(temp);
  	  return(true);
  	}
  	else
  	  return(false);
  }

  /**
   * Concatenates r to this RegExpr
   *
   * @param RegExpr - the regular expression to concatenate
   */
  public void concat(RegExpr r){
      expression=expression.concat(r.getExpression());
  }

  /**
   * Concatenates String s to this RegExpr if s is a valid regular expression
   *
   * @param s The String to concat with the expression. It must be a valid regular expression.
   * @return boolean - True if concatenation was successful.
   */
  public boolean concat(String s){
  	if(isRegExpression(s)){
  		expression=expression.concat(s);
      return(true);
  	}
  	else
  		return(false);
  }

  /**
   * Performs the Kleene Star Operation to the RegExpr
   */
  public void star(){

  	// append star to the expression
    String temp = new String(OPEN_BRACKET);
    temp=temp.concat(expression);
    temp=temp.concat(CLOSE_BRACKET);
  	temp=temp.concat(STAR);
  	expression=temp;
  }

  /**
   * Tests whether this RegExpr has an empty expression
   *
   * @return True if this RegExpr has an empty expression`
   */
  public boolean isEmpty(){
  	if(expression.length()==0)
  	  return(true);
  	else
  	  return(false);
  }

  /**
   * Tests whether the given expression is less than this one.
   * The comparision is made on the character codes of their respective expressions
   * Note: At this time only the first character is examined
   *
   * @return True if the first char of the Expression is less than the first char of r
   */
  public boolean isLessThan(RegExpr r){
  	// The type casting is necessry for the comparision of the chars
  	if(r.isEmpty())
  	  return(false);

  	if(isEmpty())
  	  return(true);

  	return((int)expression.charAt(0)<(int)r.getExpression().charAt(0));
  }

  /**
   * Tests whether the given expression is greater than this one. This the opposite
   * of isLessThan(RegExpr). The comparision is made on the character codes of their
   * respective expressions
   * Note: At this time only the first character is examined
   *
   * @return True if the first char of the Expression is greater than the first char of r
   */
  public boolean isGreaterThan(RegExpr r){
    if(r.isEmpty())
      return(true);

    if(isEmpty())
      return(false);
   	// The type casting is necessry for the comparision of the chars
  	return((int)expression.charAt(0)>(int)r.getExpression().charAt(0));
  }

  /**
   * Returns true if one RegExpr is the same as another
   * Note: They must be exactly the same
   * not the same as two regular representing the same language
   *
   * @param RegExpr - The regular expression to be tested
   * @return True if r is the same as this RegExpr
   */
  public boolean equals(Object r){
    if(!(r instanceof RegExpr))
        return false;
    if(this==r){ // if it is the same reference (ie: same instance)
      return(true);
    }

    if(r==null){// if s is null or not of the same class
      return(false);
    }

    RegExpr temp;

    try{
      temp=(RegExpr)r;
    }
    catch(ClassCastException cce){
      return(false);
    }


    if(expression.equals(temp.getExpression())){
        return true;
    }
    else{
      return(false);
    }

    //return(expression.equals(temp.getExpression()));
  }

  /**
   * Calculates the hash code of the RegExpr
   *
   * @return the int hash code of the RegExpr
   */
  public int hashCode(){
    return(expression.hashCode());
  }

  /**
   * Tests to see if RegExpr is only one letter long
   * or if it is longer than that it has no operators
   *
   * eg) 1 - is basic
   *     (1*+2)* - is complicated
   *
   * @return True if the RegExpr is only one letter long
   */
  public boolean isBasic(){
  	// check length of string
  	if(expression.length()==1)
  	  return(true);
  	else
  	  return(false);
  }

  /**
   * Tests the String s to see if it is a valid regular expression
   * A valid regual expression must have equal numbers of open and closed brackets.
   * It must also satisfy the accepted rules for regular expressions as laid out in
   * theory of computation and regular languages
   *
   * @param s The String to be tested
   * @return True if the String s is a valid regular expression
   */
  public static boolean isRegExpression(String s){
  	// iterate through the string looking for invalid form
  	// given by 2 operators beside each other or invalid bracketing

  	boolean op_fl=false;// flag to see if 2 operators are next to
  		// each other with out bracets;
    int bracket_fl=0; // to check all brackets are paired
    int i;

    // check for empty string
    if(s.length()==0 || s==null)
      return false;

    //length = s.length();
    for(i=0;i<s.length();i++){
      // count brackets
      if(s.charAt(i)=='(')
        bracket_fl++;
      if(s.charAt(i)==')')
      	bracket_fl--;

      // check operators
      if(s.charAt(i)=='+' || s.charAt(i)=='*'){
        if(op_fl)
          return(false);
        else
          op_fl=true;
      }
      else
      	op_fl=false;
    }

    if(bracket_fl!=0)
      return(false);
    else
      return(true);
  }

  /**
   * Returns the String representation of this RegExpr
   *
   * @return Returns the String representation of this RegExpr
   */
  public String toString(){
    return(expression);
  }

  /**
   * A method that will turn a regular expression into it's Postfix notation
   * so as to convert the regular expression into an Fa.
   *
   * @param String s Infix notation of the expression
   * @return Returns the Postfix representation of the expression.
   */
   private static String convertToPostfix(String s){
	 Vector opStack = new Vector();
   	 String resultingString = "";
	 int stringIterator = 0;
	 char c;


   	 opStack.add(new Character('('));//add a left parenthesis to the stack to serve as a marker to when we are complete.
   	 s+=")";//add the corresponding closing parenthesis to the end of the string we are converting.


   	 while(!opStack.isEmpty()&&stringIterator<s.length()){
   	 	c = s.charAt(stringIterator);
   	 	if(c=='('){
   	 		opStack.add(new Character(c));
   	 	}
   	 	else if(c==')'){
   	 		if(((Character)opStack.lastElement()).charValue()=='(')
   	 			opStack.remove(opStack.size()-1);
   	 		//else

			while(!opStack.isEmpty()&&((Character)opStack.lastElement()).charValue()!='('){
				char tmp = ((Character)opStack.remove(opStack.size()-1)).charValue();
				resultingString+=tmp;
			}

   	 	}
   	 	//look at the operators next.  Order of precedence - 1) star, 2)concatenation(.), 3)Union(+)
   	 	else if(c=='.'){
   	 		char tempC = ((Character)opStack.lastElement()).charValue();
   	 		while(tempC=='.'||tempC=='*'){
   	 			opStack.remove(opStack.size()-1);
   	 			resultingString+=tempC;
   	 			tempC=((Character)opStack.lastElement()).charValue();
   	 		}
   	 		opStack.add(new Character(c));
   	 	}
   	 	else if(c=='+'){
   	 		char tempC = ((Character)opStack.lastElement()).charValue();
   	 		while(tempC=='.'||tempC=='*'||tempC=='+'){
   	 			opStack.remove(opStack.size()-1);
   	 			resultingString+=tempC;
   	 			tempC=((Character)opStack.lastElement()).charValue();
   	 		}
   	 		opStack.add(new Character(c));
   	 	}
   	 	else if(c=='*'){
   	 		char tempC = ((Character)opStack.lastElement()).charValue();
   	 		while(tempC=='*'){
   	 			opStack.remove(opStack.size()-1);
   	 			resultingString+=tempC;
   	 			tempC=((Character)opStack.lastElement()).charValue();
   	 		}
   	 		opStack.add(new Character(c));
   	 	}
   	 	else{
   	 		resultingString+=c;
   	 	}

   	 	stringIterator++;
   	 }

   	 return resultingString;

   }
   
   //TODO: This method needs implementation
   /*public static Fa convertToFa(String s)
   {
   	return null;
   }*/
   

}
