/*** Giddes
 *     - Graphical Interface for Designing Discrete Event Systems
 *  S.Laurie Ricker Ph.D
 *  
 *  Developed by I. Fraser & W. Fowler 
 *           
 *  
 * reproduction without the express permission of Toba will
 * result in being taken outside and shot many times with a 
 * very tiny gun.
 ***/

//*********************************************************
// Documentation started February 8, 2002                 *
//*********************************************************

// package
package umdes;

// imports
import java.util.StringTokenizer;
import java.util.Vector;


/**
 * This Object specifies the actual State of the Plant and 
 * the respective States of the Observers of the Plant.
 */
public class StateEstimate extends State
{
  //******************************************************************
  // Private Members                                                 *
  //******************************************************************
  
  /**
   * The collection of observers of the Plant
   */
  private Vector<State> obs;
  
  //******************************************************************
  // Public Members                                                  *
  //******************************************************************
  
  /**
   * String that specifies that an Invalid Input was provided
   * If usually the message of a thrown Exception
   */
  public static final String NOT_A_STATE="Provided Input is not a State!";
  
  
  
  /**
   * Creates a new StateEstimate with the name s, and not observer
   * estimates
   *
   * @param s a String specifiying the name of the new State
   *
   * @throws Exception if the String s is not a valid State name
   */
  public StateEstimate(String s) throws Exception{
    super(s);
    obs = new Vector<State>();
  }
  
  /**
   * Creates a new StateEstimate with the specified
   * parameters.
   *
   * @param plantState the actual State of the Plant
   * @param observerStates a Vector of the Observer States. 1 for each observer
   *
   * @throws Exception if the Vector contains an object other than State
   */
  public StateEstimate(State plantState, Vector<State> observerStates) throws Exception{
    super(plantState);
    obs = observerStates;
  }
  
  /**
   * Sets the Plant State
   *
   * @param s the plant State
   */
  public void setPlantState(State s){
    name=s.toString();
  }
  
  /**
   * Adds the specified observer to the StateEstimate
   *
   * @param o the State of the observer to add
   */
  public void addObserverState(State state){
    obs.add(state);
  }

  /**
   * Deletes the specified Observer state from the StateEstimate
   *
   * @param o the observer State to remove
   *
   * @return True if the deletion was successful
   */
  public boolean deleteObserverState(State o){
    if(obs.contains(o)){
      obs.remove(o);
      return true;
    }
    return(false) ;
  }
  
  /**
   * Returns the number of observers in this estimate
   *
   * @return the number of observers
   */
  public int numObservers(){
    return(obs.size());
  } 
  
  /**
   * Converts this StateEstimate into a State object
   *
   * @return the converted State
   */
  public State convertToMAState(){
    State s=new State(this);
    return s;
  }
  
  /**
   * Creates the string representation of this estimate
   *
   * @return the string representation of this estimate
   */
  public String toString(){
    String temp=new String("{");
    int i;
    State s;
    String dump;
    
    temp+=name;
    
    
    for(i=0;i<obs.size();i++){
      if(i==0)
        temp+="|";
        
      s=(State)obs.elementAt(i);
      
      dump=s.toString();  
      
      // check to see if braces are need.
      //if(!dump.startsWith("{"))
      //  dump="{"+dump+"}";
      
      temp+=dump;
        
      if(i+1<obs.size())
        temp+="|";
    }
    
    temp+="}";
    
    return temp;
  }
  
  /**
   * Returns the plant state name of this StateEstimate
   *
   * @return the plant state name of this StateEstimate
   */
  public String getPlantStateName(){
    return(name); 
  }
  
  /**
   * Returns the plant state of this StateEstimate
   *
   * @return the plant state of this StateEstimate
   */
  public State getPlantState(){
    State s;
    try{
      s=new State(name);
    }
    catch(Exception cce){
      return null;
    }
    
    return(s); 
  }
  
  /**
   * Returns a Vector of the observer Estimates, where each Estimate
   * is a State in an Observer.
   *
   * @return Vector of the observer Estimates, where each Estimate
   * is a State in an Observer.
   */
  public Vector<State> getEstimates(){
	  Vector<State> obsClone = new Vector<State>();
	  for(State state: obs) obsClone.add(state);
	  return obsClone;
  } 
  
  /**
   * Checks the equality of the given object with this StateEstimate
   *
   * @return true if the given object is equal to this StateEstimate
   */ 
  public boolean equals(Object o){
     if(this==o)  {
      // if it is the same reference (ie: same instance)
      //System.err.println("Same reference: true");
      return(true);
    }
  
    if (o==null){
      // if s is null or not 
      //of the same class
      //System.err.println("Different classes: false");
      return(false);
    }

    StateEstimate snew; // check for Class Casting
    try 
    {
      snew=(StateEstimate)o;
    }
    catch (ClassCastException cce)
    {
      return false;
    }
    
    // check that each observer estimate is the same
    if(!name.equals(snew.toString()))   
      return(false);
      
    if(!obs.equals(getEstimates()))
      return(false);
      
      
    return(true);  
  }
  
  /**
   * Parses the given string into a StateEstimate Object
   * if the String cannot be successfully parsed an Exception is 
   * to notify of this.
   *
   * @param s the string to parse into a state estimate
   *
   * @return the generated StateEstimate
   *
   * @throws Exception if the string cannot be successfully parsed
   */
  public static StateEstimate parseIntoEstimate(String s) throws Exception{
    StringTokenizer buffer=new StringTokenizer(s,"{");
    StringBuffer temp_buff;
    String temp_string;
    

    String parse_error=new String("Parsing Error Occured"); //** like a constant
    
    Vector<String> v=new Vector<String> (); //** collection to hold parsed strings
    
    char c;
    
    if(s.indexOf(",,")>-1)
      throw new Exception(parse_error+"284");
      
    
    //*** separte string into tokens delimited on "{"
    while(buffer.hasMoreTokens()){
      temp_string=buffer.nextToken();
      temp_buff=new StringBuffer(temp_string);

      //*** Examine each token
      for(int i=0;i<temp_buff.length();i++){
        c=temp_buff.charAt(i);
        
        //** check comma placement
        if(c=='|' && (i==0 || i+1==temp_buff.length()))
          temp_buff.deleteCharAt(i);
        
        if(c=='{' && temp_buff.length()==1)
          throw new Exception(parse_error+"302");
        
        if(c=='{' && i==0)
          temp_buff.deleteCharAt(i);
        
        else if(c=='{' && i!=0){ //** open bracket doesn't at start
          throw new Exception(parse_error+"308");         
        }
        
        else if(c=='}' && i+1==temp_buff.length()){
          temp_buff.deleteCharAt(i);
        }
        
        else if(c==')' && i+1<temp_buff.length()){
          throw new Exception(parse_error+"316"); 
        }
        
        else if(i==0&&s.indexOf(temp_string) != 1){
          temp_buff.insert(0,"{");
          i++; //** make sure to skip this added char
        }
        
        else if(c=='}' && i+2==temp_buff.length()){
          //** check next char
          i++;
         
          c=temp_buff.charAt(i);
          
          if(c=='}' || c=='|'){
            temp_buff.deleteCharAt(i);
            break; //** done looking at this token
          }
          else
            throw new Exception(parse_error+"335");
        }    
        
      }
      //** made it, use temp_buff to build plant_state, observer states
      v.add(temp_buff.toString());
    }
         
    //** go through parsed strings and geneate the appriopiate states
    Vector<State> obs_states = new Vector<State>();
    State plantState=new State("DEFAULT");
    StateEstimate est;
    try{
      
      for(int i=0;i<v.size();i++){
        temp_string=(String)v.elementAt(i);
        if(i==0)
          plantState=new State(temp_string);
      
        else{
          obs_states.add(new State(temp_string));
        }
      }
      
      //** make plant state was set
      if(plantState.hasName("DEFAULT"))
        throw new Exception("ERROR");
        
      //** build estimate from states
      est=new StateEstimate(plantState,obs_states);
    }
    catch(Exception err){
      throw new Exception(parse_error);
    }
    
    return(est);
  }
  
  /**
   * Finds if any of the observer states for these two estimates
   * are equivalent and returns a boolean array where each 
   * array index is the number of the observer
   */
  public boolean[] findEquivalence(StateEstimate est2){
    Vector<State> v = est2.getEstimates();
    
    boolean result[]=new boolean[obs.size()];
    
    for(int i=0;i<obs.size();i++){
      State one = obs.elementAt(i);
      State two = v.elementAt(i);
      
      //** find if equivalent
      result[i]=one.equals(two);
    }
    
    return result;
  }
};
