/*** Giddes
 *     - Graphical Interface for Designing Discrete Event Systems
 *  S.Laurie Ricker Ph.D
 *  
 *  Developed by I. Fraser & W. Fowler 
 *                     with J.Taylor    
 *           
 *  
 * reproduction without the express permission of Toba will
 * result in being taken outside and shot many times with a 
 * very tiny gun.
 ***/

//*********************************************************
// Documentation completed January 17, 2002               *
//*********************************************************


package umdes;

/**
 * This is the Transition Object for a Finite Automata. It has a label,
 * a To State, and a From State. This object is scheduled to be deprecated
 * eventually
 *
 * @see UmdesAutomaton
 * @see State
 */

public class Transition
{
  //*************************************************************
  // private members 
  //*************************************************************
  
  /**
   * The TransitionLabel that represents the event that causes the transition
   */
  protected TransitionLabel label;
  /**
   * The State the transition starts from
   */
  protected State from;
  /**
   * The State the transition goes to
   */
  protected State to;
  
  
  private boolean legality;//////////////////// Only in GIDDES version
  
  //**************************************************************
  // public members
  //**************************************************************
  
  /**
   * Default Constructor, creates a new Transition 
   *
   * @param f The State the new Transition starts at
   * @param t The State the new Transition goes to
   * @param r The RegExpr event that causes the Transition
   */ 
  public Transition(State f, State t, RegExpr r){
   		label = new TransitionLabel(r);
  		from= new State(f);
  		to = new State(t);
  		//legality = true; /////////////////// ONLY in GIDDES version
  }
  /**
   * Creates a new Transition 
   *
   * @param f The string the new Transition starts at
   * @param t The string the new Transition goes to
   * @param r The String event that causes the Transition
   */ 
  public Transition(String f, String t, String r)throws Exception{
   		label = new TransitionLabel(r);
  		from= new State(f);
  		to = new State(t);
  		//legality =true;////////////////////////// ONly in GIDDES version
  }

  
  /**
   * Creates a new Transition 
   *
   * @param f The State the new Transition starts at
   * @param t The State the new Transition goes to
   * @param r The TransitionLabel event that causes the Transition
   */ 
  public Transition(State f, State t, TransitionLabel r){
      label = new TransitionLabel(r);
      from= new State(f);
      to = new State(t);
      //legality = true; ////////////// ONLY in GIDDES version
  }
  
  /**
   * Copy Constructor, it is preferable to use clone()
   *
   * @param t The Transition the new Transition will be a copy of.
   */
  public Transition(Transition t){
    label = new TransitionLabel(t.getLabel());
    from= new State(t.getFromState());
  	to = new State(t.getToState());
  	//legality = t.getLegality(); ////////////////////////// only in GIDDES version
  }
  
  public Object clone(){
    Transition temp=new Transition(new State(from),new State(to), new RegExpr(label));
    return(temp); 
  }
  /**
   * Set Function for From State of the Transition
   *
   * @param f the orginating State of the Transition
   */
  public void setFromState(State f){
  	from=new State(f);
  }
  
  /**
   * Set Function for To State of the Transition
   *
   * @param t the terminating State of the Transtiion
   */
  public void setToState(State t){
  	to=t;
  }
  
  /**
   * Set Function for Label of the Transition
   *
   * @param r the RegExpr label for the Transition
   */
  public void setLabel(RegExpr r){
  	label=new TransitionLabel(r);
  }
  
  /**
   * Set Function for Label of the Transition
   *
   * @param r the TransitionLabel for the Transition
   */
  public void setLabel(TransitionLabel r){
    label=r;
  }
  
  /** Accessor Function for Originating State
   *
   * @return the From State of the Transition 
   */
  public State getFromState(){
    return(new State(from));
  }
  
  /** Accessor Function for Terminating State
   *
   * @return the To State of the Transition
   */
  public State getToState(){
  	return(new State(to));
  }
  
  /** Accessor Function for Label
   *
   * @return the label of the Transition
   */
  public TransitionLabel getLabel(){
    TransitionLabel r = new TransitionLabel(label); 
  	return(r);
  }
  
  /**
   * Checks to see if the Transition contains State s as a to State
   *
   * @param s the arriving state
   * @return true if s is the To State of this transition
   */
  public boolean goesToState(State s){
  	if(s.equals(to)){
  	  return(true);
  	}
  	else
  	  return(false);
  }
  	
  /** 
   * Tests whether this Transition starts at State st 
   * and has a label (RegExpr) r
   *
   * @param st  Originating State 
   * @param r  Transition label
   * @return True if the Transition matches the given parameters
   */      
  public boolean transitionExist(State st, RegExpr r){
  	// this compares the state names of st and from and 
	  // compares the RegExpr	expressions. returns true if they 
	  // are equal
    return(st.toString().equals(from.toString()) && label.equals(new TransitionLabel(r)));
	}
  
  /**
   * Tests whether this event starts at State st
   * and has a label (regExpr) r
   *
   * @param st  Originating State 
   * @param r  Transition label
   * @return the terminating state if there is this Transition matches the parameters
   */
  public State doTransition(State st, RegExpr r){
    if(transitionExist(st,r))
		  return(to);
	  else
		  return(null);
  }
  
  /** 
   * Function test whether the Transition has only a basic (one letter)
   * regular expression for a label
   *
   * @return True is the Transition has a one letter label
   */
  public boolean isBasicTransition(){
  	return(label.isBasic());
  }
  
  /**
   * Tests whether this Transition is an Epsilon Transition
   *
   * @return True if this Transition is an Epsilon Transition
   */
  public boolean isEpsilon(){
    return(label.isEpsilon());
  }
  
  /**
   * Tests whether this Transitions is equal to a given Object
   *
   * @return True if this Transition equals Object t
   */
  public boolean equals(Object t){
    if(this==t) // if it is the same reference (ie: same instance)
      return(true);
    
    //if s is null or not of the same class
    if(t==null)
      return(false);
    
    Transition tr;
      
    try{
     tr =(Transition)t;
    }
    catch(ClassCastException cce){
      return(false);
    }
        
  	// check the individual members of Transition
  	if(!tr.getFromState().equals(from))
  	  return(false);
  	if(!tr.getToState().equals(to))
  	  return(false);
  	if(!tr.getLabel().equals(label))
  	  return(false);
  	
  	return(true);
  }
  
  /**
   * Calculates the hash code of the Transition
   *
   * @return the hash code of the Transition
   */
  public int hashCode(){
    int result=0;
    
    result=to.hashCode();
    result=47*result+from.hashCode();
    result=47*result+label.hashCode();
    
    return(result);
  }
  
  /**
   * Creates this Transition as a String output
   *
   * @return A string representation of this Transition
   */
  public String toString(){
    return(new String(from.toString()+ "->"+ to.toString()+" by "+label.toString())); 
  }
  
  public boolean getLegality() //////////////
  {
  	return legality;
  }
  
 public void setLegality(boolean legal) ////////////////
  {
  	legality = legal;
  } // These will only appear in the GIDDES version of the program.
};

		



   			
