/*** Giddes
 *     - Graphical Interface for Designing Discrete Event Systems
 *  S.Laurie Ricker Ph.D
 *  
 *  Developed by I. Fraser & W. Fowler 
 *           
 *  
 * reproduction without the express permission of Toba will
 * result in being taken outside and shot many times with a 
 * very tiny gun.
 ***/

//*********************************************************
// Documentation started February 9, 2002                 *
//*********************************************************


package umdes;


/**
 * This Object describes an event label in a Transition.
 * It usually is simply a RegExpr, but it also supports timing specifications
 * for timed automata.
 *
 * @see UmdesAutomaton
 * @see RegExpr
 * @see Transition
 */
public class TransitionLabel extends RegExpr{
  //**************************************************************
  // private Members                                             *
  //**************************************************************
  /**
   * The Timing Specification of the label
   */
  private TimeSpec specification;
  
  //**************************************************************
  // public members                                              *
  //**************************************************************
  
  /**
   * String specifying an invalid regular expression was passed into
   * a method. Usually the message of a thrown Exception
   */
  public static final String INVALID_REGEXPR="Not a valid Regular Expression";
  
  /**
   * Creates a new Label from the given String.
   * The Timing Specification is set to null by default.
   * If the String is not a valid RegExpr the expression will be
   * empty.
   *
   * @param s the String specifying the event of the label, which must be a valid
   * RegExpr or the expression will be empty
   */
  public TransitionLabel(String s){
    super(s);
    specification=null;
  }
  
  /**
   * Creates a new Label from the given RegExpr.
   * The Timing Specification is set to null by default.
   *
   * @param r the RegExpr specifying the event of the label
   */
  public TransitionLabel(RegExpr r){
    super(r);
    
    specification=null;
  }
  
  /**
   * Creates a new Label from the given String.
   * The Timing Specification is set to the provided TimeSpec.
   * If the String is not a valid RegExpr the expression will be
   * empty.
   *
   * @param s the String specifying the event of the label, which must be a valid
   * RegExpr or the expression will be empty
   */
  public TransitionLabel(String s, TimeSpec t) throws Exception{
    super(s);
    specification=t;
  }
  
  /**
   * Creates a new Label from the given RegExpr.
   * The Timing Specification is set to to the provided TimeSpec
   *
   * @param r the RegExpr specifying the event of the label
   */
  public TransitionLabel(RegExpr r, TimeSpec t){
    super(r);
    
    specification=t;
  }
  
  /**
   * Set the timing specifications of the label
   *
   * @param t the timing specification
   */
  public void setTimeSpec(TimeSpec t){
    specification=t;
  }
  
  /**
   * Returns the RegExpr event of the label
   *
   * @return the event of the label
   */
  public RegExpr getEvent(){
    return(new RegExpr(expression));
  }
  
  /**
   * Returns the Timing Specification of the label
   *
   * @return the Timing Specification of the label
   */
  public TimeSpec getTimeSpec(){
    return(specification);
  }
  
  /**
   * Creates a String Representation of this timinig specification
   *
   * @return a String Representation of this timinig specification
   */
  public String toString(){
    String temp=new String("");
    temp+=expression;
    
    if(specification!=null){
      temp+=" | ";
      temp+=specification.toString();
    }
    
    return(temp); 
  }
  
  /**
   * Tests whether the given object is equal to this TransitionLabel
   *
   * @param o the Object to test
   */
  public boolean equals(Object o){
    if(this==o){
      return true;
    }
    
    if(o==null){
      return false;
    }
    
    TransitionLabel t;
    
    try{
      t=(TransitionLabel)o;
    }
    catch(ClassCastException cce){
      return false;
    }
    
    return(t.getExpression().equals(expression));
  }
  
  /**
   * Clones this object
   */
  public Object clone(){
    TransitionLabel t;
    try{
      t=new TransitionLabel(expression,specification);
    }
    catch(Exception e){
      return null;
    }
   
    return(t); 
  }
   
}; 
   
